print("*Loaded CNG script*");

-- ==================================================
--                    Constants 
-- ==================================================
-- default values for constants;
local iMinCityNameLength = 4;
local iMaxCityNameLength = 15;
local iRenameChanceOnDistrict = 100;
local iRenameChanceOnFounded = 100;
local iRenameChanceOnConquered = 100;
local iMaxCitizenNameLength = 6;
local iIslandSizeLimit = 9;
local sOverrideCivHuman = "";
local sOverrideCivAI = "";

local iKeepOriginalCapitals = 0;

-- load options from the Advanced Setup screen
if GameConfiguration.GetValue("CNG_UseOriginalCapitals") then iKeepOriginalCapitals = 1 end;
--iKeepOriginalCapitals = GameConfiguration.GetValue("CNG_UseOriginalCapitals");
print("loaded KeepOriginalCapitals from Advanced Setup screen: " .. iKeepOriginalCapitals);
	
iRenameChanceOnDistrict = GameConfiguration.GetValue("CNG_RenameChanceOnDistrict");
print("loaded RenameChanceOnDistrict from Advanced Setup screen: " .. iRenameChanceOnDistrict);

iRenameChanceOnFounded = GameConfiguration.GetValue("CNG_RenameChanceOnFound");
print("loaded RenameChanceOnFounded from Advanced Setup screen: " .. iRenameChanceOnFounded);

iRenameChanceOnConquered = GameConfiguration.GetValue("CNG_RenameChanceOnConquest");
print("loaded RenameChanceOnConquered from Advanced Setup screen: " .. iRenameChanceOnConquered);

-- load values from user options file to replace defaults.
print("*Attempting to load CNG options from config file*");

if GameInfo.CNG_UserOptions() == nil then
	print("could not find CNG config table");
end

for Row in GameInfo.CNG_UserOptions() do
	if Row.Name == "MinCityNameLength" and Row.Value ~= nil then 
		iMinCityNameLength = Row.Value;
		print("loaded MinNameLength from config: " .. Row.Value);
	end;
	if Row.Name == "MaxCityNameLength" and Row.Value ~= nil then 
		iMaxCityNameLength = Row.Value;
		print("loaded MaxNameLength from config: " .. Row.Value);
	end;
	--[[
	if Row.Name == "RenameChanceOnDistrict" and Row.Value ~= nil then 
		iRenameChanceOnDistrict = Row.Value;
		print("loaded RenameChanceOnDistrict from config: " .. Row.Value);
	end;
	--]]
	if Row.Name == "MaxCitizenNameLength" and Row.Value ~= nil then 
		iMaxCitizenNameLength = Row.Value;
		print("loaded MaxCitizenNameLength from config: " .. Row.Value);
	end;
	if Row.Name == "IslandSizeLimit" and Row.Value ~= nil then 
		iIslandSizeLimit = Row.Value;
		print("loaded IslandSizeLimit from config: " .. Row.Value);
	end;
	--[[
	if Row.Name == "RenameChanceOnFounded" and Row.Value ~= nil then
		iRenameChanceOnFounded = Row.Value;
		print("loaded RenameChanceOnFounded from config: " .. Row.Value);
	end
	if Row.Name == "RenameChanceOnConquered" and Row.Value ~= nil then
		iRenameChanceOnConquered = Row.Value;
		print("loaded RenameChanceOnConquered from config: " .. Row.Value);
	end
	--]]
	--[[
	if Row.Name == "KeepOriginalCapitals" and Row.Value ~= nil then
		iKeepOriginalCapitals = Row.Value;
		print("loaded KeepOriginalCapitals from config: " .. Row.Value);
	end
	--]]
	
	if Row.Name == "OverrideCivHuman" and Row.Value ~= nil then
		sOverrideCivHuman = Row.Value;
		print("loaded OverrideCivHuman from config: " .. Row.Value);
	end
	if Row.Name == "OverrideCivAI" and Row.Value ~= nil then
		sOverrideCivAI = Row.Value;
		print("loaded OverrideCivAI from config: " .. Row.Value);
	end
end


-- ======================================================
--                      Event Handlers
-- ======================================================

-- table to record ownership of cities each turn, so we can check if it's changed.
local CityOwnerStartOfTurn = {};

-- at the start of every player's turn, make a record of which civ owns every city on the map.
function CNG_RecordCityOwnership()

	--clear the record table 
	CityOwnerStartOfTurn = {};
	
	--print("CNG: Checking city ownership.");
	
	-- loop through every plot on the map
	for iPlotID = 0, Map.GetPlotCount() do
		local pPlot = Map.GetPlotByIndex(iPlotID);
		
		-- if this plot is a city, record it's ownership in the table.
		if pPlot ~= nil and pPlot:IsCity() then 
			pCity = Cities.GetCityInPlot( pPlot:GetX(), pPlot:GetY() );
			--print("found " .. Locale.Lookup( pCity:GetName() ) .. " on plot " .. iPlotID .. " (" .. pPlot:GetX() .. "," .. pPlot:GetY() .. ")" );
		
			local iOwnerCivID = pCity:GetOwner();
			--print("owner civ ID: " .. iOwnerCivID);
			
			-- add ID of owner to the table, indexed by ID of the plot. (because city ID gets changed when city is captured)
			CityOwnerStartOfTurn[iPlotID] = iOwnerCivID;
			
			
			--print( "table entry number [" .. iPlotID .. "] reads: " .. CityOwnerStartOfTurn[iPlotID] );
		end
	end
end

-- call this at the start of every players turn.
Events.PlayerTurnActivated.Add(CNG_RecordCityOwnership);
-- also call it once when this script first loads. I.e. turn 0.
CNG_RecordCityOwnership();

-- auto-rename cities when they are founded
function CNG_CityFounded(iPlayerID, iCityID)
	-- get info about the city, and the player who founded it.
	local pPlayer = Players[iPlayerID];
	local pPlayerConfig = PlayerConfigurations[iPlayerID];
	local sPlayerType = pPlayerConfig:GetCivilizationTypeName();
	local pCity = pPlayer:GetCities():FindID( iCityID );
	local sOldName = Locale.Lookup( pCity:GetName() );
	
	print("CNG: city of (" .. sOldName .. ") was founded by player " .. iPlayerID .. " (" .. sPlayerType .. ")" );
	
	-- special case for liberated city states, give them back their original name
	if pPlayer:IsMajor() == false then
		print(sPlayerType .. " is a city-state. Giving it proper city-state name.");
		local sNewName = GameInfo.Civilizations[sPlayerType].Name;
		local pCity = pPlayer:GetCities():FindID( iCityID );
		pCity:SetName(sNewName);
		return false;
	end
		
	-- if override option is set, use that civType instead of the player's
	if sOverrideCivHuman ~= "" and pPlayer:IsHuman() then
		print("Override set: using " .. sOverrideCivHuman .. " for human civs");
		sPlayerType = sOverrideCivHuman;
	end
	if sOverrideCivAI ~= "" and pPlayer:IsMajor() and pPlayer:IsHuman() == false then
		print("Override set: using " .. sOverrideCivAI .. " for AI civs");
		sPlayerType = sOverrideCivAI;
	end
	
	-- abort if there is no CNG table for the civ owning the city
	if CNG_TableEntriesExistForCiv(sPlayerType) == false then
		print("No CNG table entries for " .. sPlayerType .. ", cannot generate name.");
		return false;
	end
	
	-- check our recorded data, to see who owned a city on this plot at start of turn.
	local iCityPlotID = ( Map.GetPlot( pCity:GetX(), pCity:GetY() ) ):GetIndex();
	
	
	if CityOwnerStartOfTurn[iCityPlotID] == nil then
		print ("no record of a city on plot " .. iCityPlotID .. " at the start of this turn");
	end
	local pOldOwner = Players[CityOwnerStartOfTurn[iCityPlotID]];
	
	-- if the owner of the city's plot was a different player
	-- if so, assume it's just been conquered.
	if pOldOwner ~= nil and pPlayer ~= pOldOwner then
		--print("current owner ID: " .. pCity:GetOwner() );
		--print("previous owner ID: " .. CityOwnerStartOfTurn[iCityPlotID] );
		
		print("... current owner is not the same as owner at the start of this turn.");
		local sPreviousOwner = PlayerConfigurations[CityOwnerStartOfTurn[iCityPlotID]]:GetCivilizationTypeName();
		
		-- if override option is set, use that civType instead of the new owner's
		if sOverrideCivHuman ~= "" and pOldOwner:IsHuman() then
			sPreviousOwner = sOverrideCivHuman;
		end
		if sOverrideCivAI ~= "" and pOldOwner:IsMajor() and pOldOwner:IsHuman() == false then
			sPreviousOwner = sOverrideCivAI;
		end
		
		print("original owner was: " .. sPreviousOwner );
		print("current owner is: " .. sPlayerType );
		
		-- this is not a newly founded city, so we don't want to completely rename it.
		CNG_CityConquered(pCity, sPreviousOwner, sPlayerType);
		return false;
	end
			
	-- don't rename capital cities if the option to keep them is set.
	if pCity == pPlayer:GetCities():GetCapitalCity() and iKeepOriginalCapitals == 1 then 
		print("... It is a capital City! Not renaming it.")
		return false;
	end
	
	-- random chance to ignore this event
	local iRoll = math.random(99)
	if iRoll > iRenameChanceOnFounded then
		print("CNG: Random roll prevented rename: " .. iRoll .. "/" .. iRenameChanceOnFounded);
		return false;
	end
	
	-- Since we haven't aborted for any of the above reasons, actually start generating the name now.
	
	-- Check for nearby features, resources, etc.
	local pCityPlot = Map.GetPlot( pCity:GetX(), pCity:GetY() );
	
	print("Checking for nearby resources");
	local ResourceList = CNG_GetNearbyResource(pCityPlot);
	print("Checking for nearby features");
	local FeatureList = CNG_GetNearbyFeature(pCityPlot);
	print("Checking for nearby terrains");
	local TerrainList = CNG_GetNearbyTerrain(pCityPlot);
	

	-- assemble list of words fitting those features, resource, etc.
	local ValidWordParts = CNG_CompileWordList(sPlayerType, ResourceList, FeatureList, TerrainList);
	-- generate a name from that word list.
	local sNewName = CNG_GenerateCityName(ValidWordParts, sPlayerType);

	-- if it's too long or short, roll a new one.
	if string.len(sNewName) < iMinCityNameLength or string.len(sNewName) > iMaxCityNameLength then
		print(sNewName .. " was too long/short, trying a second time.");
		sNewName = CNG_GenerateCityName(ValidWordParts, sPlayerType);
	end
	
	-- rename the city
	pCity:SetName(sNewName);
	print("Renamed " .. sOldName .. " to " .. sNewName)
	
	
	print("looking up meaning");
	local WordsContained = CNG_AnalyseName(sNewName, sPlayerType);
	
end

-- call this whenever a city is founded or changes ownership.
Events.CityInitialized.Add(CNG_CityFounded);


-- add to a city's name when it builds it's first district
function CNG_FirstDistrictAdded ( iPlayerID, iCityID, iOrderType, iProducedType, canceled, typeModifier )
	--print("A District has been built!")
	
	-- TO DO: restrict chosen words by MutuallyExclusiveGroup.
	--        shorten name if too long (drop words if space-seperated)
	
	-- get info about the city, and the player who founded it.
	local pPlayer = Players[iPlayerID];
	local pPlayerConfig = PlayerConfigurations[iPlayerID];
	local sCivType = pPlayerConfig:GetCivilizationTypeName();
	local pCity = pPlayer:GetCities():FindID( iCityID );
	local sCityName = Locale.Lookup( pCity:GetName() );

	-- if override option is set, use that civType instead of the player's
	if sOverrideCivHuman ~= "" and pPlayer:IsHuman() then
		sCivType = sOverrideCivHuman;
	end
	if sOverrideCivAI ~= "" and pPlayer:IsMajor() and pPlayer:IsHuman() == false then
		sCivType = sOverrideCivAI;
	end
	
	-- iOrderType tells us what type of thing the city is building. 0=unit, 1=building, 2=district
	-- we're only interested in districts.
	if iOrderType ~= 2 then
		--print(sCityName .. " built something, but it's not a district");
		return false;
	end
	
	-- find out what district the city is building now, and how many districts it already has.
	local sDistrictType = GameInfo.Districts[iProducedType].DistrictType;
	local iNumDistricts = pCity:GetDistricts():GetNumDistricts();
	
	-- Is this the first district that's been added to this city?
	-- note: city centre counts as a district, so we're actually looking for district number 2.
	if iNumDistricts == 2 then
		print("CNG: Player " .. iPlayerID .. " (" .. sCivType .. ") built a (" .. sDistrictType .. ") in " .. sCityName .. ". Attempting to add to name");
	
		-- ignore if there's no CNG table for this civ.
		if CNG_TableEntriesExistForCiv(sCivType) == false then
			print("No CNG table entries for " .. sCivType .. ", cannot generate name.");
			return false;
		end
		
		-- random chance to ignore this event
		local iRoll = math.random(99)
		if iRoll > iRenameChanceOnDistrict then
			print("CNG: Random roll prevented rename: " .. iRoll .. "/" .. iRenameChanceOnDistrict);
			return false;
		end
		
		-- don't rename capital cities if the option to keep them is set.
		if pCity == pPlayer:GetCities():GetCapitalCity() and iKeepOriginalCapitals == 1 then 
			print("... It is a capital City! Not renaming it.")
			return false;
		end
		
		-- find out whether the existing city name contains a MutuallyExclusiveGroup
		local sExclusiveGroup = nil;
		
		-- get a list of words in the existing name
		local WordsAlreadyInName = CNG_AnalyseName(sCityName, sCivType);
		
		-- for each of those words
		local iNumWordsAlreadyInName = CNG_GetTableLength(WordsAlreadyInName);
		
		for i = 0, iNumWordsAlreadyInName-1 do
			--print("checking word " .. WordsAlreadyInName[i] .. " for mutually exclusive flag");
			-- loop through the database to find the word.
			for Row in GameInfo.CNG_NameParts() do
				-- is it mutually exclusive?
				if Row.Text == WordsAlreadyInName[i] and Row.MutuallyExclusiveGroup ~= nil then
					sExclusiveGroup = Row.MutuallyExclusiveGroup;
				end
			end
		end		
		
		if sExclusiveGroup ~= nil then
			print(sCityName .. " contains words from MutuallyExclusiveGroup " .. sExclusiveGroup);
		end
		
		-- assemble a word list for this specific type of district.
		local WordParts = {};
		local iNumWordParts = 0;
		for Row in GameInfo.CNG_NameParts() do
			-- is part valid for this Civ?
			if Row.CivilizationPrereq == nil or Row.CivilizationPrereq == sCivType then
				-- is part valid for this district?
				if Row.DistrictPrereq == sDistrictType then
					-- is part valid for this MutuallyExclusiveGroup?
					if Row.MutuallyExclusiveGroup == nil or Row.MutuallyExclusiveGroup == sExclusiveGroup then
						-- add it to the list
						WordParts[iNumWordParts] = Row;
						print("... " .. Row.Text);
						iNumWordParts = iNumWordParts +1;
					end
				end
			end
		end	
		print("... found " .. iNumWordParts .. " valid words for " .. sDistrictType );
		
		
		--[[
		-- if we found a mutually exclusive group, strip out any potential new words it makes invalid.
		if sExclusiveGroup ~= nil then
			-- if so, strip out any potential words that don't match.
			print(sCityName .. " contains words from MutuallyExclusiveGroup " .. sExclusiveGroup);
			CNG_StripWordsFromList(WordParts, sExclusiveGroup, nil)
		end
		--]]
		
		-- if there are any valid words, pick one at random
		if iNumWordParts >0 then
			local sNewName = "";
			iRoll =  math.random(iNumWordParts);
			-- WordList[x][1] is position WordList[x][2] is text
			print("... adding: " .. WordParts[iRoll-1].Text);
			
			-- if new word is position 0 or 1, truncate the name before adding it.
			if WordParts[iRoll-1].Position ==0 then
				sNewName = WordParts[iRoll-1].Text .. CNG_TruncateName(sCityName, sCivType, 1);
			elseif WordParts[iRoll-1].Position ==1 then
				sNewName = CNG_TruncateName(sCityName, sCivType, 2) .. WordParts[iRoll-1].Text;
			-- if it's a further out position, truncate a whole space-seperated word before adding it.
			elseif WordParts[iRoll-1].Position <0 then
				sNewName = WordParts[iRoll-1].Text .. CNG_TruncateNameSpaceSeperated(sCityName, sCivType, 1);
			else
				sNewName = CNG_TruncateNameSpaceSeperated(sCityName, sCivType, 2) .. WordParts[iRoll-1].Text;
			end
			
			-- replace [space] flags with spaces
			--sNewName = string.gsub(sNewName, "%[space%]", " ");
			sNewName = CNG_ReplaceWildcards(sNewName, sCivType);
			
			-- remove any capital letters in the middle of the name
			sNewName = CNG_RemoveCamelCase(sNewName);
			
			-- rename the city
			pCity:SetName(sNewName);
			print("Renamed " .. sCityName .. " to " .. sNewName)
				
			print("looking up meaning");
			local WordsContained = CNG_AnalyseName(sNewName, sCivType);
		end
	end
end

-- Trigger this when a city builds a district
--Events.DistrictAddedToMap.Add( CNG_FirstDistrictAdded ); -- triggers when build started, but *also* whenever a savegame is loaded.
--Events.CityProductionCompleted.Add( CNG_FirstDistrictAdded ); -- triggers *twice*?? when a build is completed.
Events.CityProductionChanged.Add( CNG_FirstDistrictAdded ); -- triggers when a build is started.


-- change or add to a city's name when it's conquered by another civ.
function CNG_CityConquered (pCity, sOriginalOwner, sNewOwner)
	print("A City has been conquered!");
		
	-- random chance to ignore this event
	local iRoll = math.random(99)
	if iRoll > iRenameChanceOnConquered then
		print("CNG: Random roll prevented rename: " .. iRoll .. "/" .. iRenameChanceOnConquered);
		return false;
	end
		
	
	local sOldName = Locale.Lookup( pCity:GetName() );
	print("city of " .. sOldName .. " has changed ownership from " .. sOriginalOwner .. " to " .. sNewOwner .. " attempting to rename.")
	-- we want to add a prefix or suffix from the new civ to the old city name, truncating it if neccesary.

	
	-- Check for nearby features, resources, etc.
	local pCityPlot = Map.GetPlot( pCity:GetX(), pCity:GetY() );
	
	print("Checking for nearby resources");
	local ResourceList = CNG_GetNearbyResource(pCityPlot);
	print("Checking for nearby features");
	local FeatureList = CNG_GetNearbyFeature(pCityPlot);
	print("Checking for nearby terrains");
	local TerrainList = CNG_GetNearbyTerrain(pCityPlot);
	--print("Checking for districts owned");
	--local DistrictsList = 
	
	-- build a list of valid words from the new owner's wordlist.
	local WordList = CNG_CompileWordList(sNewOwner, ResourceList, FeatureList, TerrainList);
	
	-- pick one of these words at random, and add it to the old name.
	local sNewName = "";
	local iRoll = 0;
	local iNumWordsInList = CNG_GetTableLength(WordList)-1;
	
	if iNumWordsInList > 0 then
		iRoll =  math.random(iNumWordsInList);
			
			print("... adding: " .. WordList[iRoll-1].Text);
			
			-- if new word is position 0 or 1, truncate name (according to old owner's words) before adding it.
			if WordList[iRoll-1].Position ==0 then
				sNewName = WordList[iRoll-1].Text .. CNG_TruncateName(sOldName, sOriginalOwner, 1);
			elseif WordList[iRoll-1].Position ==1 then
				sNewName = CNG_TruncateName(sOldName, sOriginalOwner, 2) .. WordList[iRoll-1].Text;
			-- if it's a further out position, truncate a whole space-seperated word before adding it.
			elseif WordList[iRoll-1].Position <0 then
				sNewName = WordList[iRoll-1].Text .. CNG_TruncateNameSpaceSeperated(sOldName, sOriginalOwner, 1);
			else
				sNewName = CNG_TruncateNameSpaceSeperated(sOldName, sOriginalOwner, 2) .. WordList[iRoll-1].Text;
			end
			
			-- replace wildcards
			sNewName= CNG_ReplaceWildcards(sNewName, sCivType);
			
			-- remove any capital letters in the middle of the name
			sNewName = CNG_RemoveCamelCase(sNewName);
			
			-- rename the city
			pCity:SetName(sNewName);
			print("Renamed " .. sOldName .. " to " .. sNewName)
				
			print("looking up meaning");
			local WordsContained = CNG_AnalyseName(sNewName, sNewOwner, sOriginalOwner );
			
	end
end




-- ======================================================
--                     Name Generation
-- ======================================================

-- compile a list of word-parts valid for this civ and location
function CNG_CompileWordList (sCivType, ResourceList, FeatureList, TerrainList, DistrictList)
	
	local iNumWordParts = 0;
	local WordParts = {};

	print("Assembling list of word parts")
	
	-- loop through all entries in NameParts table
	for Row in GameInfo.CNG_NameParts() do
		-- is part valid for this Civ?
		if Row.CivilizationPrereq == nil or Row.CivilizationPrereq == sCivType then
			-- is part valid for this Terrain?
			if Row.TerrainPrereq == nil or CNG_ExistsInList(Row.TerrainPrereq, TerrainList) then
				-- is part valid for this Feature?
				if Row.FeaturePrereq == nil or CNG_ExistsInList(Row.FeaturePrereq, FeatureList) then
					-- is part valid for this Resource?
					if Row.ResourcePrereq == nil or CNG_ExistsInList(Row.ResourcePrereq, ResourceList) then
						-- if it has a district prereq, ignore it for now
						if Row.DistrictPrereq == nil then
							-- add that row from the database to the list.
							WordParts[iNumWordParts] = Row;
							iNumWordParts = iNumWordParts +1;
						end
				
					end
				end
			end
		end
	end
	
	print("There are " .. iNumWordParts .. " valid word parts")
	
	-- return the completed list
	return WordParts;
	
end

-- function will return a name assembled from the list of name-parts fed to it.
function CNG_GenerateCityName(WordList, sCivType)
	if WordList == nil then
		print("invalid word list passed to CNG_GenerateCityName");
		return;
	end
	
	
	-- string that will be returned later 
	local sNewName = "";
	local iRoll = 0;
	local iNumWordsInList = CNG_GetTableLength(WordList);
	
	--seperate lists for words of each Position value
	local WordsPos0 = {};
	local iCountPos0 = 0;
	local WordsPos1 = {};
	local iCountPos1 = 0;
	local WordsPos2 = {};
	local iCountPos2 = 0;
	local WordsPos3 = {};
	local iCountPos3 = 0;
	local WordsPosNeg1 = {};
	local iCountPosNeg1 = 0;
	local WordsPosNeg2 = {};
	local iCountPosNeg2 = 0;
	local WordsPosNeg3 = {};
	local iCountPosNeg3 = 0;
	
	
	print("starting to Generate Name")
	print("... there are " .. iNumWordsInList .. " words available");
	-- Note: remember, lua tables begin at index 1, not index 0.
	
	-- word list is a table of tables. Each entry contains a whole row from the database.

	
	-- first, we need to sort the input list by position, creating smaller lists for each position.
	for i = 0, (iNumWordsInList-1) do
		-- is this word for position 0?
		if WordList[i].Position == 0 then
			-- add it to the pos0 list
			WordsPos0[iCountPos0] = WordList[i];
			--print(WordsPos0[iCountPos0] .. " is pos 0");
			iCountPos0 = iCountPos0 +1;
		end
		-- is this word for pos1?
		if WordList[i].Position == 1 then
			-- add it to the pos1 list
			WordsPos1[iCountPos1] = WordList[i];
			--print(WordsPos1[iCountPos1] .. " is pos 1");
			iCountPos1 = iCountPos1 +1;
		end	
		-- is this word for pos2?
		if WordList[i].Position == 2 then
			-- add it to the pos1 list
			WordsPos2[iCountPos2] = WordList[i];
			--print(WordsPos2[iCountPos2] .. " is pos 2");
			iCountPos2 = iCountPos2 +1;
		end
		-- is this word for pos3?
		if WordList[i].Position == 3 then
			-- add it to the pos3 list
			WordsPos3[iCountPos3] = WordList[i];
			--print(WordsPos3[iCountPos3] .. " is pos 3");
			iCountPos3 = iCountPos3 +1;
		end
		-- and now the negative positions
		-- is this word for pos-1?
		if WordList[i].Position == -1 then
			-- add it to the posNeg1 list
			WordsPosNeg1[iCountPosNeg1] = WordList[i];
			--print(WordsPosNeg1[iCountPosNeg1] .. " is pos -1");
			iCountPosNeg1 = iCountPosNeg1 +1;
		end
		-- is this word for pos-2?
		if WordList[i].Position == -2 then
			-- add it to the posNeg2 list
			WordsPosNeg2[iCountPosNeg2] = WordList[i];
			--print(WordsPosNeg2[iCountPosNeg2] .. " is pos -2");
			iCountPosNeg2 = iCountPosNeg2 +1;
		end
		-- is this word for pos-3?
		if WordList[i].Position == -3 then
			-- add it to the posNeg3 list
			WordsPosNeg3[iCountPosNeg3] = WordList[i];
			--print(WordsPosNeg3[iCountPosNeg3] .. " is pos -3");
			iCountPosNeg3 = iCountPosNeg3 +1;
		end
	end
	
	
	
	-- pick a random pos0
	iRoll =  math.random(iCountPos0);
	sNewName = sNewName .. WordsPos0[iRoll-1].Text;
	print("Pos 0: " .. WordsPos0[iRoll-1].Text);
	
	-- Is the first word we've chosen part of a mutually exclusive group?.
	if WordsPos0[iRoll-1].MutuallyExclusiveGroup ~= nil then
		print(WordsPos0[iRoll-1].Text .. " is in mutualy exclusive group " .. WordsPos0[iRoll-1].MutuallyExclusiveGroup)
	end
	
	-- strip out any words from lists that are no longer valid. Either because of exclusive groups or identical text.
	WordsPos1, iCountPos1 = CNG_StripWordsFromList(WordsPos1, WordsPos0[iRoll-1].MutuallyExclusiveGroup, WordsPos0[iRoll-1].Text);
	WordsPos2, iCountPos2 = CNG_StripWordsFromList(WordsPos2, WordsPos0[iRoll-1].MutuallyExclusiveGroup, WordsPos0[iRoll-1].Text);
	WordsPos3, iCountPos3 = CNG_StripWordsFromList(WordsPos3, WordsPos0[iRoll-1].MutuallyExclusiveGroup, WordsPos0[iRoll-1].Text);
	WordsPosNeg1, iCountPosNeg1 = CNG_StripWordsFromList(WordsPosNeg1, WordsPos0[iRoll-1].MutuallyExclusiveGroup, WordsPos0[iRoll-1].Text);
	WordsPosNeg2, iCountPosNeg2 = CNG_StripWordsFromList(WordsPosNeg2, WordsPos0[iRoll-1].MutuallyExclusiveGroup, WordsPos0[iRoll-1].Text);
	WordsPosNeg3, iCountPosNeg3 = CNG_StripWordsFromList(WordsPosNeg3, WordsPos0[iRoll-1].MutuallyExclusiveGroup, WordsPos0[iRoll-1].Text);
	
	
	-- start adding other parts, chance to add each is weighted by current length of name.
	local iNumAdditionalParts = 0;
	-- pos1
	iRoll =  math.random(iCountPos1 + string.len(sNewName) );
	if iRoll < iCountPos1 then
		sNewName = sNewName .. WordsPos1[iRoll-1].Text;
		iNumAdditionalParts = iNumAdditionalParts +1;
		print("Pos 1: " .. WordsPos1[iRoll-1].Text);
		
		-- strip out any words from lists that are no longer valid. Either because of exclusive groups or identical text.
		WordsPos2, iCountPos2 = CNG_StripWordsFromList(WordsPos2, WordsPos1[iRoll-1].MutuallyExclusiveGroup, WordsPos1[iRoll-1].Text);
		WordsPos3, iCountPos3 = CNG_StripWordsFromList(WordsPos3, WordsPos1[iRoll-1].MutuallyExclusiveGroup, WordsPos1[iRoll-1].Text);
		WordsPosNeg1, iCountPosNeg1 = CNG_StripWordsFromList(WordsPosNeg1, WordsPos1[iRoll-1].MutuallyExclusiveGroup, WordsPos1[iRoll-1].Text);
		WordsPosNeg2, iCountPosNeg2 = CNG_StripWordsFromList(WordsPosNeg2, WordsPos1[iRoll-1].MutuallyExclusiveGroup, WordsPos1[iRoll-1].Text);
		WordsPosNeg3, iCountPosNeg3 = CNG_StripWordsFromList(WordsPosNeg3, WordsPos1[iRoll-1].MutuallyExclusiveGroup, WordsPos1[iRoll-1].Text);
	
	end
	-- pos-1
	iRoll =  math.random(iCountPosNeg1 + string.len(sNewName) );
	if iRoll < iCountPosNeg1 then
		sNewName = WordsPosNeg1[iRoll-1].Text .. sNewName;
		iNumAdditionalParts = iNumAdditionalParts +1;
		print("Pos -1: " .. WordsPosNeg1[iRoll-1].Text);
		
		-- strip out any words from lists that are no longer valid. Either because of exclusive groups or identical text.
		WordsPos2, iCountPos2 = CNG_StripWordsFromList(WordsPos2, WordsPosNeg1[iRoll-1].MutuallyExclusiveGroup, WordsPosNeg1[iRoll-1].Text);
		WordsPos3, iCountPos3 = CNG_StripWordsFromList(WordsPos3, WordsPosNeg1[iRoll-1].MutuallyExclusiveGroup, WordsPosNeg1[iRoll-1].Text);
		WordsPosNeg2, iCountPosNeg2 = CNG_StripWordsFromList(WordsPosNeg2, WordsPosNeg1[iRoll-1].MutuallyExclusiveGroup, WordsPosNeg1[iRoll-1].Text);
		WordsPosNeg3, iCountPosNeg3 = CNG_StripWordsFromList(WordsPosNeg3, WordsPosNeg1[iRoll-1].MutuallyExclusiveGroup, WordsPosNeg1[iRoll-1].Text);
	
	end
	
	-- pos2
	iRoll =  math.random(iCountPos2 + string.len(sNewName) );
	if iRoll < iCountPos2 then
		sNewName = sNewName .. WordsPos2[iRoll-1].Text;
		iNumAdditionalParts = iNumAdditionalParts +1;
		print("Pos 2: " .. WordsPos2[iRoll-1].Text);
		
		-- strip out any words from lists that are no longer valid. Either because of exclusive groups or identical text.
		WordsPos3, iCountPos3 = CNG_StripWordsFromList(WordsPos3, WordsPos2[iRoll-1].MutuallyExclusiveGroup, WordsPos2[iRoll-1].Text);
		WordsPosNeg2, iCountPosNeg2 = CNG_StripWordsFromList(WordsPosNeg2, WordsPos2[iRoll-1].MutuallyExclusiveGroup, WordsPos2[iRoll-1].Text);
		WordsPosNeg3, iCountPosNeg3 = CNG_StripWordsFromList(WordsPosNeg3, WordsPos2[iRoll-1].MutuallyExclusiveGroup, WordsPos2[iRoll-1].Text);
	
	end
	-- pos-2
	iRoll =  math.random(iCountPosNeg2 + string.len(sNewName) );
	if iRoll < iCountPosNeg2 then
		sNewName = WordsPosNeg2[iRoll-1].Text .. sNewName;
		iNumAdditionalParts = iNumAdditionalParts +1;
		print("Pos -2: " .. WordsPosNeg2[iRoll-1].Text);
		
		-- strip out any words from lists that are no longer valid. Either because of exclusive groups or identical text.
		WordsPos3, iCountPos3 = CNG_StripWordsFromList(WordsPos3, WordsPosNeg2[iRoll-1].MutuallyExclusiveGroup, WordsPosNeg2[iRoll-1].Text);
		WordsPosNeg3, iCountPosNeg3 = CNG_StripWordsFromList(WordsPosNeg3, WordsPosNeg2[iRoll-1].MutuallyExclusiveGroup, WordsPosNeg2[iRoll-1].Text);
	
	end
	
	-- pos3
	iRoll =  math.random(iCountPos3 + string.len(sNewName) );
	if iRoll < iCountPos3 then
		sNewName = sNewName .. WordsPos3[iRoll-1].Text;
		iNumAdditionalParts = iNumAdditionalParts +1;
		print("Pos 3: " .. WordsPos3[iRoll-1].Text);
		
		-- strip out any words from lists that are no longer valid. Either because of exclusive groups or identical text.
		WordsPosNeg3, iCountPosNeg3 = CNG_StripWordsFromList(WordsPosNeg3, WordsPos3[iRoll-1].MutuallyExclusiveGroup, WordsPos3[iRoll-1].Text);
	
	end
	-- pos-3
	iRoll =  math.random(iCountPos3 + string.len(sNewName) );
	if iRoll < iCountPosNeg3 then
		sNewName = WordsPosNeg3[iRoll-1].Text ..  sNewName;
		iNumAdditionalParts = iNumAdditionalParts +1;
		print("Pos -3: " .. WordsPosNeg3[iRoll-1].Text);
	end
	
	-- if no additional parts at all were chosen, try to add a pos 1 one more time.
	if iNumAdditionalParts == 0 then
		print("only one word chosen, second attempt to choose a Pos 1")
		iRoll =  math.random(iCountPos1 + string.len(sNewName) );
		if iRoll < iCountPos1 then
			sNewName = sNewName .. WordsPos1[iRoll-1].Text;
			iNumAdditionalParts = iNumAdditionalParts +1;
			print("Pos 1: " .. WordsPos1[iRoll-1].Text);
		end
	end
	
	-- replace any wildcards, such as [space] or [citizen]
	sNewName= CNG_ReplaceWildcards(sNewName, sCivType);
	
	-- remove any capital letters in the middle of the name
	sNewName = CNG_RemoveCamelCase(sNewName);
	
	print("Generated the name " .. sNewName);
	return sNewName;
end


-- takes a complete city name as input, returns a list of words it contains, and a list of their meanings.
function CNG_AnalyseName (sString, sCivType, sCivType2)
	print("Checking for meanings in the name " .. sString)
	
	local ResultsTable = {};
	local ResultsTableMeanings = {};
	local iNumResults = 0;
	
	local sCheckedWord = "";
	local iStartPoint = 1;
	local sUnknownWord = "";
		
	-- step through the string from left to right
	while iStartPoint < string.len(sString) do
		-- using this char as a start point, step back from right to left.
		for iEndPoint=string.len(sString), iStartPoint, -1 do
			local sCheckedWord = string.sub(sString, iStartPoint, iEndPoint);
			
			-- does this substring match a word in the database table?
			local sMeaning = CNG_LookupWordMeaning(sCheckedWord, sCivType);
			-- if a second civ type was passed, check their word list too.
			if sMeaning == false and sCivType2 ~= nil then 
				sMeaning = CNG_LookupWordMeaning(sCheckedWord, sCivType2);
			end
			if (sMeaning ~= false) then
				-- when we find a matching word, first check if there was an unknown (not matching anything) word preceeding it
				if sUnknownWord ~= "" then
					-- add the unknown word to the results table 
					ResultsTable[iNumResults] = sUnknownWord;
					ResultsTableMeanings[iNumResults] = "Unknown";
					iNumResults = iNumResults +1;
					print(sUnknownWord .. ": Unknown")
					-- reset unknown word string
					sUnknownWord = "";
				end
				-- add the matching word to the table of results.
				ResultsTable[iNumResults] = sCheckedWord;
				ResultsTableMeanings[iNumResults] = sMeaning;
				iNumResults = iNumResults +1;
				print(sCheckedWord .. ": " .. sMeaning);
				-- then jump startpoint to the end of this word
				iStartPoint = iEndPoint + 1;
			end
			
			-- if we somehow stepped all the way from endpoint to startpoint and didn't find a matching word...
			if iEndPoint == iStartPoint then
				-- add this character to the "unknown word" string
				sUnknownWord = sUnknownWord .. string.sub(sString, iStartPoint, iEndPoint);
				-- and move on to the next character.
				iStartPoint = iStartPoint+1;
			end	
			
			-- if stepped all the way from startpoint to the end of the name, without finding a meaning...
			if iStartPoint >= string.len(sString) and sUnknownWord ~= "" then
				-- add the unknown word to the results table 
					ResultsTable[iNumResults] = sUnknownWord;
					ResultsTableMeanings[iNumResults] = "Unknown";
					iNumResults = iNumResults +1;
					print(sUnknownWord .. ": Unknown")
			end
			
		end
	end
	
	return ResultsTable;
end

-- looks through the CNG_NameParts table for the input string, returns the matching meaning.
function CNG_LookupWordMeaning(sString, sCivType)
	-- replace any spaces in input string with format used in the table.
	sString = string.lower( string.gsub(sString, " ", "%[space%]") );
	
	if sString == nil or sString == "" then
		return false;
	end
	
	-- loop through Text entries in NameParts table, looking for this string.
	for Row in GameInfo.CNG_NameParts() do
	
		if Row.Text == nil or Row.Text == "" then
			print("Error: found a CNG table entry with no Text")
		else
			-- adjust Row.Text to ignore any [citizen] wildcards
			local sAdjustedRowText = string.lower( string.gsub(Row.Text, "%[citizen%]", "") );
			
			-- check if this row in the database matches the input word
			-- also check input string with a space at start or end, just in case.
			if sAdjustedRowText ~= "" and sAdjustedRowText == sString or sAdjustedRowText == "[space]" .. sString or sAdjustedRowText == sString .. "[space]" then
				-- if a sCivType argument was passed, then filter by this civ only
				if sCivType == nil or Row.CivilizationPrereq == sCivType then
					if Row.Meaning ~= nil and Row.Meaning ~= "" then
						return Row.Meaning;
					end
				end		
			end
		end
	end
	
	-- Once we've tried all words in the wordList, check if this is a citizen name
	local sFullCitizenName = CNG_IsCitizenName(sString, sCivType);
	
	if sFullCitizenName ~= false then
		return sFullCitizenName .. " (personal name)";
	end
	
	-- if we couldn't find a meaning for this word, return false.
	return false;	
end


-- accepts a city name as input, returns it with the first or last meaningful word removed.
function CNG_TruncateName(sName, sCivType, iWhichEnd)

	-- get a list of individual words contained in the name.
	local WordsContained = CNG_AnalyseName(sName, sCivType);
	local iNumWords = CNG_GetTableLength(WordsContained);
	
	-- if there's only one word in the name then don't cut anything out.
	if iNumWords < 2 then
		print("CNG: Can't truncate " .. sName .. ", because it only contains a single word.");
		return sName;
	else
		-- if there's more than one word, we need to remove one. Either first or last, depending on sWhichEnd argument.
		local sNewName = "";
		
		if iWhichEnd == 1 then
			print("truncating start of " .. sName);
			-- add all words except the first one on the list
			for i=1, iNumWords-1 do
				sNewName = sNewName .. WordsContained[i]; 
			end
			return sNewName;
		elseif iWhichEnd == 2 then
			print("truncating end of " .. sName);
			-- add all words except the last one on the list
			for i=0, iNumWords-2 do
				sNewName = sNewName .. WordsContained[i]; 
			end
			return sNewName;
		else
			print("CNG error: invalid iWhichEnd argument to CNG_TruncateName")
			return sName;
		end
		
	end
end


-- accepts a city name as input, returns it with the first or last meaningful word removed.
function CNG_TruncateNameSpaceSeperated(sName, sCivType, iWhichEnd)

	-- get a list of individual words contained in the name.
	local WordsContained = CNG_AnalyseName(sName);
	local iNumWords = CNG_GetTableLength(WordsContained);
	
	-- if there's only one word in the name then don't cut anything out.
	if iNumWords < 2 then
		print("CNG: Can't truncate " .. sName .. ", because it only contains a single word.");
		return sName;
	else
		-- if there's more than one word, we need to remove one. Either first or last, depending on sWhichEnd argument.
		local sNewName = "";
				
		if iWhichEnd == 2 then
			print("truncating space-seperated start of " .. sName);
			-- loop through from left to right.
			for i=0, iNumWords-1 do
				-- does this word begin or end with a space or a dash?
				if string.find(WordsContained[i], "[%s&-].") ~= nil then -- if it begins with a space...
					-- don't add this word, instead return what we have so far.
					return sNewName;
				elseif string.find(WordsContained[i], ".[%s&-]") ~= nil then -- if it ends with a space...
					-- add this word, then return what we have so far. 
					sNewName = sNewName .. WordsContained[i]; 
					return sNewName;
				else -- if it neither starts nor ends with a space...
					-- add this word, then carry on.
					sNewName = sNewName .. WordsContained[i]; 
				end
			end
			-- if we somehow got all the way through, return the full name.
			return sNewName;
			
		elseif iWhichEnd == 1 then
			print("truncating space-seperated end of " .. sName);
			-- loop through from right to left.
			for i= iNumWords-1, 0, -1 do
				-- does this word begin or end with a space?
				if string.find(WordsContained[i], "[%s%-].") ~= nil then -- if it begins with a space...
					-- add this word, then return what we have so far. 
					sNewName = WordsContained[i] .. sNewName; 
					return sNewName;
				elseif string.find(WordsContained[i], ".[%s%-]") ~= nil then -- if it ends with a space...
					-- don't add this word, instead return what we have so far.
					return sNewName;
				else -- if it neither starts nor ends with a space...
					-- add this word, then carry on.
					sNewName = WordsContained[i] .. sNewName; 
				end
			end
			-- if we somehow got all the way through, return the full name.
			return sNewName;
			
		else
			print("CNG error: invalid iWhichEnd argument to CNG_TruncateNameSpaceSeperated")
			return sName;
		end
		
	end
end


function CNG_RemoveCamelCase (sString)
	local sNewString = "";

	local CharsToCheck = {" ", "-"};
	
	--print("removing CamelCase from: " .. sString);
	
	-- first convert the whole thing to lowercase
	sString = string.lower(sString); 
	--print("... to lowercase: " .. sString)
	
	-- read from the string one character at a time
	for i=1, string.len(sString) do
		local sCharacter = string.sub(sString, i, i);
		-- if this is the first character, or is preceeded by a space, then make it uppercase
		if i == 1 or CNG_ExistsInList(string.sub(sString, i-1, i-1), CharsToCheck) then
			sCharacter = string.upper(sCharacter);
		end
		-- add this character to the new string
		sNewString = sNewString .. sCharacter;
	end
	 
	-- return the converted string.
	--print("... with words capitalised: " .. sNewString);
	
	return sNewString;
end



-- accepts a string and returns it with special [in brackets] parts replaced.
function CNG_ReplaceWildcards (sString, sCivType)

	-- replace any occurence of [space] with a space
	sString = string.gsub(sString, "%[space%]", " ");
	
	-- replace any double spaces with a single space
	sString = string.gsub(sString, "  ", " ");
	
	-- does the string contain [citizen] ?
	if string.find(sString, "%[citizen%]") ~= nil then
		-- look through CivilizationCitizenNames table for names appropriate for this civ.
		local CitizenNames = {}
		local iNumCitizenNames = 0;
	
		for Row in GameInfo.CivilizationCitizenNames() do
			-- if this citizen name is for this civ, add it to the list.
			if Row.CivilizationType == sCivType then			
				iNumCitizenNames = iNumCitizenNames +1;
				-- get the localised version of the TXT_KEY and add it to the list
				CitizenNames[iNumCitizenNames] = Locale.Lookup(Row.CitizenName);
			end
		end
		print("... choosing a citizen name from " .. iNumCitizenNames .. " possibilities");
		
		-- pick one of the citizen names at random, and replace the wildcard with it.
		local iRoll = math.random(iNumCitizenNames);
		local sChosenCitizenName = CitizenNames[iRoll];
		
		print("... chosen: " .. sChosenCitizenName);
		
		-- if the citizen name is too long, chop off the final third of it.
		local iNameLength = string.len(sChosenCitizenName);
		if iNameLength > iMaxCitizenNameLength then
			sChosenCitizenName = string.sub(sChosenCitizenName, 0, math.floor(iNameLength * 0.7) )
			print("... but that's too long, shortening to: " .. sChosenCitizenName);
		end
				
		
		
		sString = string.gsub(sString, "%[citizen%]", sChosenCitizenName);
		
	end
	
	-- return the finished string
	return sString;
end

-- accepts a string as input, checks if it matches a full or truncated version of a citizen name
-- returns the full name if so.
function CNG_IsCitizenName (sString, sCivType)
	--print("checking whether " .. sString .. " is a citizen name")
	
	-- change the input string to lowercase
	sString = string.lower(sString);
	
	-- loop through the table of citizen names.
	for Row in GameInfo.CivilizationCitizenNames() do
		-- if this citizen name for this civ?
		if Row.CivilizationType == sCivType then			
			-- get the localised version of the TXT_KEY. In lowercase.
			local sName = string.lower(Locale.Lookup(Row.CitizenName) );
			
			-- does the input string match this name?
			-- also check it with a space before or after, just in case.
			if sString == sName or sString == sName .. "[space]" or sString == "[space]" .. sName then
				--print(sString .. " matches " .. sName);
				return Locale.Lookup(Row.CitizenName);
			end
			
			-- long names get truncated by a third when generating city-names
			if math.floor(string.len(sName)) > iMaxCitizenNameLength then
			
				local sNameTruncated = string.sub( sName, 0, math.floor(string.len(sName) * 0.7) );
				-- does the input match a truncated version of this name?
				-- also check with a space before and after.
				if sString == sNameTruncated or sString == sNameTruncated .. "[space]" or sString == "[space]" .. sNameTruncated then
					return Locale.Lookup(Row.CitizenName)
				end
			
			end
		end
	end
	
	-- if the string doesn't match any name
	return false;
end

-- accepts a list of words and criteria for removal. Returns the same list with words matching those criteria removed.
function CNG_StripWordsFromList(WordList, sMutualyExlusiveGroup, sMatchingText)
	
	if WordList == nil then
		print("Error: invalid WordList passed to CNG_StripWordsFromList");
	end
	

	local iNumWords = CNG_GetTableLength(WordList);
	--print("list contains " .. iNumWords .. " words");
	local NewList = {};
	local iNumWordsNewList = 0;
	
	--print("attempting to strip words from list");
	
	-- loop through all words in the input list
	for i = 0, iNumWords-1 do
		--print("checking word " .. WordList[i].Text);
		
		-- if a mutualy exclusive group was passed, exclude words from different groups.		
		-- never xclude words which have no exclusive group.
		if WordList[i].MutuallyExclusiveGroup == nil or (sMutualyExlusiveGroup ~= nil and sMutualyExlusiveGroup == WordList[i].MutuallyExclusiveGroup) then
		
			-- if no MatchingText argument was passed, then there is no further reason to exclude this word.
			if sMatchingText == nil then
				-- add word to the list.
				NewList[iNumWordsNewList] = WordList[i];
				iNumWordsNewList = iNumWordsNewList +1
			else
				-- exclude words whose text is identical to the matching text. (even if there's a space before or after it)
				local sOldString = string.lower(WordList[i].Text);
				local sNewString = string.lower(sMatchingText);
				if sOldString ~= sNewString and sOldString ~= "[space]" .. sNewString and sOldString ~= sNewString .. "[space]" then
					-- if it's not been excluded, add it to the new list.
					NewList[iNumWordsNewList] = WordList[i];
					iNumWordsNewList = iNumWordsNewList +1
				end
			end
		end
	end
	
	--print ("list of " .. iNumWords .. " stripped down to " .. iNumWordsNewList .. " words");
	
	-- return the new list, and it's length.
	return NewList, iNumWordsNewList;
end

-- ===============================================
--                 Utility Functions
-- ===============================================
function CNG_TableEntriesExistForCiv(sCivType)

	-- loop through all entries in NameParts table
	for Row in GameInfo.CNG_NameParts() do
		-- as soon as we find an entry for this civ, break the loop and return true
		if Row.CivilizationPrereq == sCivType then
			return true;
		end
	end
	-- if we never found an entry for this civ, return false
	return false
end

-- accepts a string and a table of strings as arguments, returns true if at least one item in the table matches the tested string
function CNG_ExistsInList(sItem, List)
	if sItem == nil or List == nil then
		print("CNG_ExistsInList: invalid arguments")
		return false;
	end
	
	for index, item in pairs(List) do
		if item == sItem then
			return true;
		end
	end

	return false;
end

-- accepts a table, returns the number of entries in that table
function CNG_GetTableLength (Table)
	local iCount = 0;

	if Table ~= nil then
		for Entry in pairs(Table) do 
			iCount = iCount + 1;
		end
	end

	return iCount;
end




-- ===============================================
--                 Map Checking Functions
-- ===============================================


-- function accepts a plot as input, returns list of all nearby Resources
function CNG_GetNearbyResource (pPlot)
	local ResourceList = {};
	local iCount = 0;
	local iPlotX = pPlot:GetX();
	local iPlotY = pPlot:GetY();

	-- check plot 
	--print("checking centre plot");
	if pPlot:GetResourceType() ~= -1 then
		-- add resource to list if any present
		ResourceList[iCount] = pPlot:GetResourceType();
		print("center plot has " .. GameInfo.Resources[pPlot:GetResourceType()].ResourceType );
			
		iCount = iCount +1;
	end

	-- check all adjacent plots
	--print("checking adjacent plots");
	for iDirection = 0,5 do
		local pCheckedPlot = Map.GetAdjacentPlot(iPlotX, iPlotY, iDirection);
		if pCheckedPlot:GetResourceType() ~= -1 then
			-- add resource to list of any present
			ResourceList[iCount] = GameInfo.Resources[pCheckedPlot:GetResourceType()].ResourceType;
			print("... plot " .. iDirection .. " has " .. GameInfo.Resources[pCheckedPlot:GetResourceType()].ResourceType );
			iCount = iCount +1;
		end
	end

	-- return the list
	return ResourceList;

end

-- function accepts a plot as input, returns list of all nearby Features
function CNG_GetNearbyFeature (pPlot)
local FeatureList = {};
	local iCount = 0;
	local iPlotX = pPlot:GetX();
	local iPlotY = pPlot:GetY();

	-- check central plot 
	if pPlot:GetFeatureType() ~= -1 then
		-- add feature to list if any present
		FeatureList[iCount] = GameInfo.Features[pPlot:GetFeatureType()].FeatureType;
		iCount = iCount +1;
	end
	if pPlot:IsRiver() then
		FeatureList[iCount] = "FEATURE_RIVER";
		iCount = iCount +1;
		print("centre plot has FEATURE_RIVER");
	end
	if pPlot:IsHills() then
		FeatureList[iCount] = "FEATURE_HILL";
		iCount = iCount +1;
		print("centre plot has FEATURE_HILL");
		-- if a hill is adjacent to water, then it's also a cliff
		if pPlot:IsCoastalLand() then
			FeatureList[iCount] = "FEATURE_CLIFF";
			iCount = iCount +1;
		print("centre plot has FEATURE_CLIFF");
		end	
	end
	if CNG_IsIsland(pPlot) then
		FeatureList[iCount] = "FEATURE_ISLAND";
		iCount = iCount +1;
		print("centre plot is on a FEATURE_ISLAND");
	end


	-- check all adjacent plots
	--print("checking adjacent plots");
	for iDirection = 0,5 do
		local pCheckedPlot = Map.GetAdjacentPlot(iPlotX, iPlotY, iDirection);
		if pCheckedPlot:GetFeatureType() ~= -1 then
			-- add resource to list of any present
			FeatureList[iCount] = GameInfo.Features[pCheckedPlot:GetFeatureType()].FeatureType;
			iCount = iCount +1;
			print("... plot " .. iDirection .. " has "  .. GameInfo.Features[pCheckedPlot:GetFeatureType()].FeatureType);
		end

		-- The following are not returned by GetFeatureType, so check for them seperately
		if pCheckedPlot:IsHills() then
			FeatureList[iCount] = "FEATURE_HILL";
			iCount = iCount +1;
			print("... plot " .. iDirection .. " has FEATURE_HILL");
			-- if a hill is adjacent to water, then it's also a cliff
			if pCheckedPlot:IsCoastalLand() then
				FeatureList[iCount] = "FEATURE_CLIFF";
				iCount = iCount +1;
			print("... plot " .. iDirection .. " has FEATURE_CLIFF");
			end	
		end
		if pCheckedPlot:IsLake() then
			FeatureList[iCount] = "FEATURE_LAKE";
			iCount = iCount +1;
			print("... plot " .. iDirection .. " has FEATURE_LAKE");
		end
		if pCheckedPlot:IsMountain() then
			FeatureList[iCount] = "FEATURE_MOUNTAIN";
			iCount = iCount +1;
			print("... plot " .. iDirection .. " has FEATURE_MOUNTAIN");
		end
	end

	return FeatureList;

end

-- function accepts a plot as input, returns list of all nearby Terrain Types
function CNG_GetNearbyTerrain (pPlot)
	local TerrainList = {};
	local iCount = 0;
	local iPlotX = pPlot:GetX();
	local iPlotY = pPlot:GetY();

	-- check centre plot 
	TerrainList[iCount] = GameInfo.Terrains[pPlot:GetTerrainType()].TerrainType;
	print("... centre plot has " .. GameInfo.Terrains[pPlot:GetTerrainType()].TerrainType);
	iCount = iCount +1;

	-- check all adjacent plots
	--print("checking adjacent plots");
	for iDirection = 0,5 do
		local pCheckedPlot = Map.GetAdjacentPlot(iPlotX, iPlotY, iDirection);
		-- add resource to list of any present
		TerrainList[iCount] = GameInfo.Terrains[pCheckedPlot:GetTerrainType()].TerrainType;
		iCount = iCount +1;
		print("... plot " .. iDirection .. " has "  .. GameInfo.Terrains[pCheckedPlot:GetTerrainType()].TerrainType);

		-- The following terrain types are not noticed by GetTerrainType, so do seperate checks for them
	
	

	end

	return TerrainList;
end


-- accepts a plot, returns true if it's part of a small island (less than 9 hexes of land surrounded by water)
function CNG_IsIsland (pPlot)
	--print("Checking whether plot (" .. pPlot:GetX() .. "," .. pPlot:GetY() .. ") is on an island");
	
	local iNumPlots = pPlot:GetArea():GetPlotCount();
	
	--print("area has " .. iNumPlots .. " plots");
	
	if iNumPlots <= iIslandSizeLimit then
		--print("is an island");
		return true;
	else
		--print("not an island");
		return false;
	end
end
	
--[[
function CNG_IsIsland (pPlot)
	print("*** Checking whether plot (" .. pPlot:GetX() .. "," .. pPlot:GetY() .. ") is on an island ***");
		
	-- TO DO: this function is not working very well, returning a lot of false positives. Find a better way?
	
	-- make a table to contain a list of plots we've checked.
	local pPlotsList = {};
	local iNumPlotsOnList = 1;
	
	-- start with just the input plot.
	pPlotsList[iNumPlotsOnList] = pStartingPlot;
	
	-- start stepping through adjacent plots, until we have counted enough of them.
	for i=0,iIslandSizeLimit+1 do
		
		-- consider the last plot added to the list.
		local pPlot = pPlotsList[iNumPlotsOnList];
		local iPlotX = pPlot:GetX();
		local iPlotY = pPlot:GetY();
		--print("considering plot (" .. iPlotX .. "," .. iPlotY .. ")");
		
		-- look at each of the plots adjacent to this plot.
		for iDirection = 0,5 do
			local pAdjacentPlot = Map.GetAdjacentPlot(iPlotX, iPlotY, iDirection);
			
			-- if plot is on the edge of the map, it might return nil instead of a valid adjacent plot
			if pAdjacentPlot ~= nil then
				-- if the adjacent plot is already on the list, ignore it.
				if CNG_ExistsInList(pAdjacentPlot, pPlotsList) == false then

					-- if it's not on the list, check what sort of terrain it is.
					local sTerrainType = GameInfo.Terrains[pAdjacentPlot:GetTerrainType()].TerrainType;
					--print("adjacent plot: " .. iDirection .. " (" .. pAdjacentPlot:GetX() .. "," .. pAdjacentPlot:GetY() .. ") is: " .. sTerrainType);
					
					-- if the adjacent plot is not water, then it must be land. So add it to list.
					if sTerrainType ~= "TERRAIN_COAST" and sTerrainType ~= "TERRAIN_OCEAN" then
						iNumPlotsOnList = iNumPlotsOnList +1;
						pPlotsList[iNumPlotsOnList] = pAdjacentPlot;
						break;
					end	
					
				end
			end
		end	
	end
	
	-- if we've completed the loop without finding more too many land plots, then this is an island.
	if iNumPlotsOnList < iIslandSizeLimit then
		--print("city is on an island, containing " .. iNumPlotsOnList .. " land plots in total.");
		--print("*****************************************************");
		return true;
	else
		--print("not an island");
		--print("*****************************************************");
		return false;
	end
	

end
--]]

